/*
 * Copyright (c) 2019 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Fri Jan 11 2019
 */
#include "acgps.h"

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include <err.h>
#include <fcntl.h>
#include <poll.h>
#include <stdio.h>
#include <unistd.h>

#include "codels.h"

static const int keepalive_min = 60;	/* seconds */
static const int keepalive_max = 300;	 /* seconds */

static int *			rtkfd(struct gps_rtk_conn_s *rtk_conn,
                                        int domain);
static struct rtk_addrinfo_s *	lookaddr(struct rtk_addrinfo_s *list,
                                        const struct sockaddr *addr,
                                        size_t addrlen);
static struct rtk_addrinfo_s *	addr2info(const struct sockaddr *addr,
                                        size_t addrlen);
static char *			addr2name(struct sockaddr *addr, char *buf,
                                        size_t len);


/* --- gpsrtk_conn_open ---------------------------------------------------- */

int
gpsrtk_conn_open(gps_rtk_conn_s *rtk_conn, uint16_t port)
{
  static const int domain[] = { AF_INET, AF_INET6 };

  union { struct sockaddr_in addr4; struct sockaddr_in6 addr6; } addr;
  size_t i, s;
  int fd;

  gpsrtk_conn_close(rtk_conn);

  for(i = 0; i < 2; i++) {
    fd = socket(domain[i], SOCK_DGRAM, 0);
    if (fd < 0) { warn("socket"); continue; }

    switch(domain[i]) {
      case AF_INET:
        addr.addr4.sin_family = domain[i];
        addr.addr4.sin_port = htons(port);
        addr.addr4.sin_addr.s_addr = INADDR_ANY;
        s = sizeof(addr.addr4);
        break;

      case AF_INET6:
        setsockopt(fd, IPPROTO_IPV6, IPV6_V6ONLY, &(int){1}, sizeof(int));
        addr.addr6.sin6_family = domain[i];
        addr.addr6.sin6_port = htons(port);
        addr.addr6.sin6_flowinfo = 0;
        addr.addr6.sin6_addr = in6addr_any;
        addr.addr6.sin6_scope_id = 0;
        s = sizeof(addr.addr6);
        break;

      default:
        s = 0;
    }

    if (bind(fd, (struct sockaddr *)&addr, s) < 0) {
      warn("bind %d", port);
      close(fd);
      continue;
    }

    fcntl(fd, F_SETFL, fcntl(fd, F_GETFL, 0) | O_NONBLOCK);

    *rtkfd(rtk_conn, domain[i]) = fd;
  }

  if (rtk_conn->fd4 < 0 && rtk_conn->fd6 < 0)
    return (errno = EINVAL);

  warnx("serving rtk data on port %d", port);
  return 0;
}


/* --- gpsrtk_server_open -------------------------------------------------- */

int
gpsrtk_server_open(const char host[64], uint16_t port, gps_rtk_conn_s *rtk_conn)
{
  (void)rtk_conn;
  struct addrinfo hints, *res, *res0;
  struct rtk_addrinfo_s *serv;
  struct sockaddr_storage addr;
  socklen_t addrlen;
  struct pollfd fd;
  char sport[6];
  ssize_t n;
  int i, s;

  /* resolve host */
  memset(&hints, 0, sizeof(hints));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_DGRAM;
  hints.ai_flags = AI_NUMERICSERV;
  snprintf(sport, sizeof(sport), "%d", port);
  s = getaddrinfo(host, sport, &hints, &res0);
  if (s) {
    warnx("%s:%d: %s", host, port, gai_strerror(s));
    return -1;
  }

  /* check for existing connection */
  for (res = res0; res; res = res->ai_next) {
    serv = lookaddr(rtk_conn->servers, res->ai_addr, res->ai_addrlen);
    if (serv) {
      warnx("%s:%d: already listening", host, port);
      errno = 0;
      goto done;
    }
  }

  /* create new connection (retry 3 times at most each address) */
  for(i = 0; i < 3; i++) {
    for (res = res0; res; res = res->ai_next) {
      fd.fd = *rtkfd(rtk_conn, res->ai_family);
      if (fd.fd < 0) continue;

      /* send empty datagram and expect quick empty reply */
      if (sendto(fd.fd, NULL, 0, 0, res->ai_addr, res->ai_addrlen))
        continue;

      fd.events = POLLIN;
      s = poll(&fd, 1, 250/*ms*/);
      if (s <= 0) {
        char hostname[256];
        warnx("%s: timeout contacting server",
              addr2name(res->ai_addr, hostname, sizeof(hostname)));
        continue;
      }

      for(s = 1; s;) {
        addrlen = sizeof(addr);
        n = recvfrom(fd.fd, NULL, 0, 0, (struct sockaddr *)&addr, &addrlen);
        if (n < 0) break;

        if (addrlen != res->ai_addrlen) continue;
        if (memcmp(&addr, res->ai_addr, addrlen)) continue;
        s = 0;
      }
      if (s) {
        warnx("%s:%d: no data from server", host, port);
        continue;
      }

      /* create */
      serv = addr2info(res->ai_addr, res->ai_addrlen);
      if (!serv) goto done;

      serv->keepalive.tv_sec += keepalive_min;
      serv->next = rtk_conn->servers;
      rtk_conn->servers = serv;
      warnx("%s:%d: listening for rtk data", host, port);
      errno = 0;
      goto done;
    }
  }

  errno = EINVAL;
done:
  freeaddrinfo(res0);
  return errno;
}


/* --- gpsrtk_recv --------------------------------------------------------- */

int
gpsrtk_recv(gps_rtk_conn_s *rtk_conn, sequence_octet *rtk_in)
{
  (void)rtk_in;

  int fd[] = { rtk_conn->fd4, rtk_conn->fd6 };
  struct rtk_addrinfo_s *host;
  struct sockaddr_storage addr;
  socklen_t addrlen;
  char data[256];
  ssize_t n;
  size_t i;

  for(i = 0; i < sizeof(fd)/sizeof(fd[0]); i++) {
    if (fd[i] < 0) continue;

    while(1) {
      addrlen = sizeof(addr);
      n = recvfrom(fd[i], data, sizeof(data), 0,
                   (struct sockaddr *)&addr, &addrlen);
      if (n < 0) break;

      switch (n) {
        case 0: /* new client || keepalive */

          /* look for known server */
          host = lookaddr(
            rtk_conn->servers, (struct sockaddr *)&addr, addrlen);
          if (host) {
            /* server ack: ignore */
            break;
          }

          /* ack client by sending back 0 bytes */
          if (sendto(fd[i], NULL, 0, 0, (struct sockaddr *)&addr, addrlen) < 0)
            break;

          /* look for known client */
          host = lookaddr(
            rtk_conn->clients, (struct sockaddr *)&addr, addrlen);
          if (host) {
            /* already recorded: update keepalive */
            gettimeofday(&host->keepalive, NULL);
            host->keepalive.tv_sec += keepalive_max;
            break;
          }

          /* record client */
          host = addr2info((struct sockaddr *)&addr, addrlen);
          if (!host) break;
          host->keepalive.tv_sec += keepalive_max;
          host->next = rtk_conn->clients;
          rtk_conn->clients = host;

          warnx("%s: pushing rtk data",
                addr2name((struct sockaddr *)&addr, data, sizeof(data)));
          break;

        default: /* rtk data */

          /* look for known server */
          host = lookaddr(
            rtk_conn->servers, (struct sockaddr *)&addr, addrlen);
          if (!host) {
            warnx("%s: unregistered server pushing rtk data",
                  addr2name((struct sockaddr *)&addr, data, sizeof(data)));
            break;
          }

          if (rtk_in->_maximum < rtk_in->_length + n)
            if (genom_sequence_reserve(rtk_in, rtk_in->_length + n)) {
              warn("discarding rtk data");
              break;
            }

          memcpy(&rtk_in->_buffer[rtk_in->_length], data, n);
          rtk_in->_length += n;
          break;
      }
    }
  }

  return 0;
}


/* --- gpsrtk_send --------------------------------------------------------- */

int
gpsrtk_send(gps_rtk_conn_s *rtk_conn, sequence_octet *rtk_out)
{
  struct rtk_addrinfo_s *clnt;
  char hostname[64];
  ssize_t n;
  int fd;

  if (rtk_out->_length == 0) return 0;

  for(clnt = rtk_conn->clients; clnt; clnt = clnt->next) {
    fd = *rtkfd(rtk_conn, clnt->addr.ss_family);
    if (fd < 0) continue;

    n = sendto(fd, rtk_out->_buffer, rtk_out->_length, 0,
               (struct sockaddr *)&clnt->addr, clnt->addrlen);
    if (n < 0)
      warn("%s", addr2name(
             (struct sockaddr *)&clnt->addr, hostname, sizeof(hostname)));
  }

  rtk_out->_length = 0;
  return 0;
}


/* --- gpsrtk_keepalive ---------------------------------------------------- */

int
gpsrtk_keepalive(gps_rtk_conn_s *rtk_conn)
{
  struct rtk_addrinfo_s *host, *next, *prev;
  struct timeval tv;
  char name[64];
  ssize_t n;
  int fd;

  gettimeofday(&tv, NULL);

  /* send keepalive packet to servers */
  for(host = rtk_conn->servers; host; host = host->next) {
    if (host->keepalive.tv_sec > tv.tv_sec) continue;
    if (host->keepalive.tv_sec == tv.tv_sec &&
        host->keepalive.tv_usec > tv.tv_usec) continue;

    fd = *rtkfd(rtk_conn, host->addr.ss_family);
    if (fd < 0) continue;

    n = sendto(fd, NULL, 0, 0,
               (struct sockaddr *)&host->addr, host->addrlen);
    if (n < 0)
      warn("%s", addr2name(
             (struct sockaddr *)&host->addr, name, sizeof(name)));

    host->keepalive.tv_sec = tv.tv_sec + keepalive_min;
    host->keepalive.tv_usec = tv.tv_usec;
  }

  /* clear dead clients */
  for(host = prev = rtk_conn->clients; host; host = next) {
    next = host->next;
    if (host->keepalive.tv_sec > tv.tv_sec ||
        (host->keepalive.tv_sec == tv.tv_sec &&
         host->keepalive.tv_usec > tv.tv_usec)) {
      prev = host;
      continue;
    }

    warnx("%s: dropping client",
          addr2name((struct sockaddr *)&host->addr, name, sizeof(name)));

    if (prev == rtk_conn->clients)
      rtk_conn->clients = next;
    else
      prev->next = next;
    free(host);
  }

  return 0;
}


/* --- gpsrtk_conn_close --------------------------------------------------- */

void
gpsrtk_conn_close(gps_rtk_conn_s *rtk_conn)
{
  if (rtk_conn->fd4 >= 0) close(rtk_conn->fd4);
  if (rtk_conn->fd6 >= 0) close(rtk_conn->fd6);

  rtk_conn->fd4 = rtk_conn->fd6 = -1;
}


/* --- rtkfd --------------------------------------------------------------- */

static int *
rtkfd(struct gps_rtk_conn_s *rtk_conn, int domain)
{
  static int badfd;

  switch(domain) {
    case AF_INET: return &rtk_conn->fd4;
    case AF_INET6: return &rtk_conn->fd6;
  }

  badfd = -1;
  return &badfd;
}


/* --- lookaddr ------------------------------------------------------------ */

static struct rtk_addrinfo_s *
lookaddr(struct rtk_addrinfo_s *list,
         const struct sockaddr *addr, size_t addrlen)
{
  struct rtk_addrinfo_s *host;

  for(host = list; host; host = host->next) {
    if (host->addrlen != addrlen) continue;
    if (memcmp(&host->addr, addr, addrlen)) continue;

    /* match */
    break;
  }

  return host;
}


/* --- addr2name ----------------------------------------------------------- */

static struct rtk_addrinfo_s *
addr2info(const struct sockaddr *addr, size_t addrlen)
{
  struct rtk_addrinfo_s *ai;

  ai = malloc(sizeof(*ai));
  if (!ai) return NULL;

  memcpy(&ai->addr, addr, addrlen);
  ai->addrlen = addrlen;
  gettimeofday(&ai->keepalive, NULL);
  ai->next = NULL;

  return ai;
}


/* --- addr2name ----------------------------------------------------------- */

static char *
addr2name(struct sockaddr *addr, char *buf, size_t len)
{
  switch(addr->sa_family) {
    case AF_INET:
      inet_ntop(
        addr->sa_family, &((struct sockaddr_in *)addr)->sin_addr,
        buf, len);
      snprintf(buf + strlen(buf), len - strlen(buf), ":%d",
               ntohs(((struct sockaddr_in *)addr)->sin_port));
      break;

    case AF_INET6:
      inet_ntop(
        addr->sa_family, &((struct sockaddr_in6 *)addr)->sin6_addr,
        buf, len);
      snprintf(buf + strlen(buf), len - strlen(buf), ":%d",
               ntohs(((struct sockaddr_in6 *)addr)->sin6_port));
      break;

    default:
      snprintf(buf, len, "<unknown>");
  }

  return buf;
}


/* --- gpsrtk_match_rtcm --------------------------------------------------- */

int
gpsrtk_is_rtcm(const uint8_t *raw, uint16_t *type, size_t *length)
{
  /* basic detection of RTCM3 preamble */

  if (raw[0] == 0xd3) {
    *length = (((raw[1] & 0x3) << 8) | raw[2]) + 6 /* header + CRC */;
    if (type) *type = (raw[3] << 4) | (raw[4] >> 4);
    return 3;
  }

  /* basic detection of RTCM2 preamble, disregarding parity */

  /* The two MSB bits of each bytes are '01', and the preamble is contained in
   * the first 8 bits of the first 30 bits word, stuffed with 6-of-8
   * format. Preamble should be 0x66 (or 0x99 if the previous message bits
   * imply bit reversing).
   *                           byte 1             byte 2
   * bit (1 is MSB):      . . 6 5 4 3 2 1    . . . . . . 8 7
   * preamble 0x66:       0 1 1 0 0 1 1 0    0 1         0 1
   * XORed preamble 0x99: 0 1 0 1 1 0 0 1    0 1         1 0
   */
  uint8_t preamble = 0;
  if (raw[0] == 0x66 && (raw[1] & 0xc3) == 0x41)
    preamble = 0x66;
  else if (raw[0] == 0x59 && (raw[1] & 0xc3) == 0x42)
    preamble = 0x99;
  else
    return 0;

  /* Length of message is in bits 17-21 of second 30 bits word.
   * Must be XORed if preamble was 0x99.
   *
   *                           byte 8             byte 9
   * bit (1 is MSB):      . . 2 1 . . . .    . . . . . 5 4 3
   * frame length:        0 1 l l . . . .    0 1 . . . l l l
   */
  /* if ((raw[7] & 0xc0) != 0x40 || (raw[8] & 0xc0) != 0x40) return 0; */
  *length = 0;
  if (raw[7] & 0x10) *length |= 0x10;
  if (raw[7] & 0x20) *length |= 0x8;
  if (raw[8] &  0x1) *length |= 0x4;
  if (raw[8] &  0x2) *length |= 0x2;
  if (raw[8] &  0x4) *length |= 0x1;
  if (raw[4] & 0x20) *length = *length ^ 0x1f;
  *length = (*length + 2) * 5;

  /* if ((raw[2] & 0xc0) != 0x40) return 0; */
  if (type) {
    *type = 0;
    if (raw[1] &  0x4) *type |= 0x20;
    if (raw[1] &  0x8) *type |= 0x10;
    if (raw[1] & 0x10) *type |= 0x8;
    if (raw[1] & 0x20) *type |= 0x4;
    if (raw[2] &  0x1) *type |= 0x2;
    if (raw[2] &  0x2) *type |= 0x1;
    if (preamble == 0x99) *type = *type ^ 0x3f;
  }

  return 2;
}

int
gpsrtk_match_rtcm(const uint8_t *raw, size_t max,
                  const uint8_t **match, uint16_t *type, size_t *length)
{
  int s;

  if (max < 8) return 0;

  for(*match = raw; *match < raw + max - 8; (*match)++) {
    s = gpsrtk_is_rtcm(*match, type, length);
    if (!s) continue;

    if (*match + *length <= raw + max) return s;
  }

  return 0;
}


/* --- gpsrtk_copy_rtcm_stats ---------------------------------------------- */

int
gpsrtk_copy_rtcm_stats(gps_rtcm_stats_s *dst, gps_rtcm_stats_s *src)
{
  if (dst->sent._maximum < src->sent._length &&
      genom_sequence_reserve(&dst->sent, src->sent._length))
    return 1;
  memcpy(dst->sent._buffer, src->sent._buffer,
         src->sent._length * sizeof(*src->sent._buffer));
  dst->sent._length = src->sent._length;

  if (dst->recv._maximum < src->recv._maximum &&
      genom_sequence_reserve(&dst->recv, src->recv._length))
    return 1;
  memcpy(dst->recv._buffer, src->recv._buffer,
         src->recv._length * sizeof(*src->recv._buffer));
  dst->recv._length = src->recv._length;

  dst->inbytes = src->inbytes;
  dst->inpkts = src->inpkts;
  dst->outbytes = src->outbytes;
  dst->outpkts = src->outpkts;
  dst->dropbytes = src->dropbytes;
  dst->droppkts = src->droppkts;

  return 0;
}
